#ifndef CGAL_BEZIER_CURVE_2_H
#define CGAL_BEZIER_CURVE_2_H

#include "Types.h"
#include "Helpers.h"
#include "Curve_2.h"
#include "Parametric_curve_2.h"
#include "Subdivision_curve_2.h"
#include "X_FO.h"
#include "Y_FO.h"
#include "DeCasteljau_FO.h"
#include "Horner_FO.h"
#include "Bernstein_basis_FO.h"


// Template class Bezier_curve_2 implements a Bezier curve in 2D 
// defined with P(t) = Sum(Pk * Bk,n(t)), where Pk are the control points
// and Bk,n the Bernstein polynomials av degree n
// Template parameters:
//       NumType_      number type for coordinates
//       A_            the type of the algorithm which calculates for a 
//                     given parameter value t the appropriate point on
//                     the Bezier curve
// Inherits from Parametric_curve_2 and Subdivision_curve_2
template <class NumType_, class A_ = DeCasteljau_FO<NumType_> >
class Bezier_curve_2 : public Parametric_curve_2<NumType_, 
                                                 X_FO<NumType_, A_*>*, 
		                                 Y_FO<NumType_, A_*>* > , 
                       public Subdivision_curve_2<NumType_> {
public:
    typedef NumType_ NumType;
    typedef A_ A;
    typedef X_FO<NumType, A* > Bezier_x;
    typedef Y_FO<NumType, A* > Bezier_y;
    typedef Parametric_curve_2<NumType, Bezier_x*, Bezier_y*> P_curve; 
    typedef Subdivision_curve_2<NumType> S_curve;
        
    

    // Bezier_curve_2: void -> Bezier_curve_2    
    // Effect: contructs a empty Bezier_curve_2 object whose parameter is
    //         in the default interval [LOW_BOUND, UP_BOUND]
    // contructor without parameter    
    Bezier_curve_2() : P_curve(), S_curve(), 
	alg(&contr_points), bezier_x(&alg), bezier_y(&alg) {	
	P_curve::init(&bezier_x, &bezier_y, LOW_BOUND, UP_BOUND);
    }

    // Bezier_curve_2: Points, NumType, NumType -> Bezier_curve_2 
    // Parameters:
    //    contr_pts   the control points of the Bezier curve
    //    a_          the low bound of t's range, t in [a_,b_] 
    //    b_          the up bound of t's range, t in [a_,b_] 
    // Effect: contructs a Bezier_curve_2 object with control points 
    //         contr_pts whose parameter t is in [a_,b_], 
    //         default [LOW_BOUND, UP_BOUND]   
    Bezier_curve_2(const Points & contr_pts,  
		   NumType a_=LOW_BOUND, NumType b_=UP_BOUND): S_curve(contr_pts), 
	           alg(&contr_points, a_, b_),
                   bezier_x(&alg), 
                   bezier_y(&alg) {	
	P_curve::init(&bezier_x, &bezier_y, a_, b_);	
	initial_approximation();
    };

    // Bezier_curve_2: Bezier_curve_2 -> Bezier_curve_2 
    // Parameters:
    //    bezier   Bezier curve
    // Effect: contructs a Bezier_curve_2 object which is a copy of bezier   
    Bezier_curve_2(const Bezier_curve_2 & bezier) : 
    	S_curve(bezier.control_points()), 
	alg(&(bezier.control_points()), bezier.get_lower_bound(), 
		 bezier.get_upper_bound()),
	bezier_x(&alg), 
        bezier_y(&alg) {
	P_curve::init(bezier.get_x_func(), bezier.get_y_func(), 
		      bezier.get_lower_bound(), bezier.get_upper_bound());
	initial_approximation();
    };

    virtual ~Bezier_curve_2() {	
    };
    
    // initial_approximation: unsigned -> Lin_approximant
    // Parameters:
    //     num       the number of segments in the linear approximant, 
    //               default value NUMBER    
    // Pre: num > 1
    // Effect: calculates a initial piecewise linear approximant for 
    //         this curve
    // Overwrites Parametric_curve_2::initial_approximation()
    void initial_approximation(unsigned num=NUMBER) {
	Lin_approximant tmp(contr_points.begin(), contr_points.end(), 
			    a, b);      
	lin_approx = tmp;
	eps = initial_epsilon();
    };

    // get_name: void -> char*
    // Returns: the string "Bezier Curve"
    char * get_name() const {
	return "Bezier Curve";
    }; 

    // deCasteljau: NumType -> Point
    // Parameters:
    //   t    the parameter value
    // Pre: a<=t<=b
    // Returns: a point on this Bezier curve calculated with the de Casteljau 
    //          algorithm at t
    Point deCasteljau(NumType t) const {
	Points contr_pts(contr_points.begin(), contr_points.end());
	DeCasteljau_FO<NumType> de_Casteljau_alg(&contr_pts, a, b);
	return de_Casteljau_alg(t);
    }

    // horner: NumType -> Point
    // Parameters:
    //   t    the parameter value
    // Pre: a<=t<=b
    // Returns: a point on the this Bezier curve calculated with 
    //          the Horner algorithm at t
    Point horner(NumType t) const {
	Points contr_pts(contr_points.begin(), contr_points.end());
	Horner_FO<NumType> horner_alg(&contr_pts, a, b);
	return horner_alg(t);	
    }
    
    // subdivide: NumType -> Points, Points
    // Parameters:
    //   l    the vector for the control points of the left subdivision curve
    //   r    the vector for the control points of the right subdivision curve
    //   m    the parameter value at wich the curve will be subdivided, 
    //        default value (b+a)/2
    // Pre: a<=m<=b
    // Effect: calculates the control points of two bezier curves resulting
    //         after subdividing this Bezier curve at m. The left curve is 
    //         defined in [a, m] the right in [m, b]. They have the same
    //         degree as the original curve
    void subdivide(Points& l, Points& r, NumType m=0.5) const{
	if(m==0.5 && (m<a || m>b)) m=(b+a)/(NumType) 2;
	assert(a<=m && m<=b);
	int n = degree();
	Points pts = contr_points;	
	l.resize(n + 1);	
	r.resize(n + 1);       
	l[0] = pts[0];	
	r[n] = pts[n];
	for(int j = (n-1); j>=0; j--) {
	    for(int k = 0; k<=j; k++) {		
	      //pts[k]=(1-m)*pts[k]+m*pts[k+1]
	        pts[k]=between_p_q(pts[k], pts[k+1], m);		
	    }
	    l[n-j] = pts[0];
	    r[j] = pts[j];
	}
    }
    
    // within_tolerance: NumType -> boolean
    // Parameter:
    //    epsilon   the tolerance, dafault value EPSILON
    // Returns:
    //    true    if the distance from each one of the internal control 
    //            points to line joining the first and the last control
    //            points is <= epsilon 
    //    false   otherwise
    // Determines whether a Bezier curve can be approximated to within 
    // tolerance by the straight line joining its first and last control 
    // points
    bool within_tolerance(NumType epsilon=EPSILON) const {
	int n = degree();
	NumType e_2 = power(epsilon, 2);
	Segment s(contr_points[0], contr_points[n]);
	for(int i=1; i<n; i++) {	    
	    if(squared_distance(contr_points[i], s) > e_2) 
		return false;
        }
	return true;
    }
    
    // rendering: NumType -> Points
    // Parameters:
    //   epsilon   the tolerance, dafault value EPSILON
    //   points    vector of points. The linking polygon is gives a
    //             piecewise linear approximation of  this curve within 
    //             a tolerance epsilon
    // Effect: performs rekursiv subdivision until the curve can be
    //         approximated to whitin tolerance by the line segment
    //         joining the first and the last control points
    void rendering(Points& points, NumType epsilon=EPSILON) const {
	if(within_tolerance(epsilon)) {
	    if(points.empty() || points[points.size()-1]!=contr_points[0])
		points.push_back(contr_points[0]);
	    points.push_back(contr_points[degree()]);	    
	    return;
	}
	else {		
	    subdivide(left_pts, right_pts);
	    Bezier_curve_2 * left_bezier = new Bezier_curve_2(left_pts);
	    Bezier_curve_2 * right_bezier = new Bezier_curve_2(right_pts);
	    (* left_bezier).rendering(points, epsilon);
	    (* right_bezier).rendering(points, epsilon);
	    delete left_bezier;
	    delete right_bezier;
	}
    }

    // rendering_n_times: unsigned -> Points
    // Parameters:
    //   n         times the rendering procedure will be executed
    //   points    vector of points wich approximates this curve
    // Effect: performes n rekursiv subdivisions
    void rendering_n_times(Points& points, unsigned n) const {
        if(n==0) {	    
	    if(points.empty() || points[points.size()-1]!=contr_points[0])
		points.push_back(contr_points[0]);
	    points.push_back(contr_points[degree()]);
	    return;
	}
	else {	       
	    subdivide(left_pts, right_pts);	    	    	    
	    Bezier_curve_2 * left_bezier = new Bezier_curve_2(left_pts);
	    Bezier_curve_2 * right_bezier = new Bezier_curve_2(right_pts);    
	    (* left_bezier).rendering_n_times(points, (n-1));	    
	    (* right_bezier).rendering_n_times(points, (n-1));	    
	    delete left_bezier;	    
	    delete right_bezier;	    
	}
    }
    
    // derivative: unsigned -> Bezier_curve_2, bool
    // Papameters:    
    //    r          compute the r-th derivative of this Bezier curve,
    //               default value r=1
    //    deriv      contains at the end the r-th derivative (Bezier 
    //               curve of degree n-r), if it exists  
    // Pre:     r>0 
    // Returns: true, if the r-th discrete derivative exists (r<degree)
    //          false otherwise
    bool derivative(Bezier_curve_2 & deriv, unsigned r=1) const {
	unsigned n = degree();
	assert(r > 0);
	if(n < r) return false; 
	NumType coef = n;
	for(unsigned i=n-1; i>(n-r); i--) {
	    coef=coef*i;
	}	
	Points q;
	discrete_differences(q, r);
	PointList tmp;
	for(unsigned i = 0; i<q.size(); i++) {
	    tmp.push_back( CGAL::ORIGIN + (q[i] -CGAL::ORIGIN)*coef);
	}		
	deriv.set_control_points(tmp);         
	return true;
    }
    

    // evaluate_derivative_at: NumType, int -> Vector, bool
    // Papameters:
    //    t      the parameter at which the derivative is computed
    //    r      compute the r-th derivative of the curve, default r=1
    //    deriv   will contain the r-th derivative of the curve at t, if 
    //            it exists
    // Returns: true, if the r-th derivative at t exists
    //          false otherwise
    // Pre:  a<=t<=b    
    // Overwrites Parametric_curve_2::evaluate_derivative_at()
    bool evaluate_derivative_at(NumType t, Vector & deriv, 
			     unsigned r=DEGREE) const {
	assert(a<=t && b>=t);
	assert(r>=1);	
	Points tmp;
	Bezier_curve_2 bez_deriv(tmp, a, b);
	if(!derivative(bez_deriv, r)) 
	    return false;
	deriv = (bez_deriv.calculate_point_at(t) - CGAL::ORIGIN);
	return true;
    }
       
    // max_subdiv_level: NumType -> unsigned
    // Parameter:
    //    epsilon   the tolerance, dafault value 0.001
    // Returns:  the number subdivision levels requiered to anssure that
    //           every segment of this curve will be approximated to within 
    //           epsilon by the straight line joining the curve end points 
    // Uses the Wang's Formula
    unsigned max_subdiv_level(NumType epsilon=EPSILON) const {
	if(degree()<=1) return 0;
	Bezier_curve_2 deriv;
	derivative(deriv, 2);
	// the control_points of the second derivative 
	// are P''=P_k+2 - 2*P_k+1 + P_k
	Points pts = deriv.control_points();
	//max_2= (Max|P_k+2 - 2*P_k+1 + P_k|)^2
	NumType max_2 = (pts[0] - CGAL::ORIGIN)*(pts[0] - CGAL::ORIGIN);
	for(unsigned i=1; i<pts.size(); i++) {
	    NumType tmp_2 = (pts[i] - CGAL::ORIGIN)*(pts[i] - CGAL::ORIGIN);	    
	    if(max_2<tmp_2) max_2=tmp_2;
	}
	//L>=log4(sqrt(max_2)/8e)
	double arg = _sqrt(max_2) / _to_float((8*epsilon));
	//log4(arg)=ln(arg)/ln(4)=ln(arg)/2*ln(2)=ln(arg)/2*M_LN2	
	unsigned res = (unsigned) ceil((log(arg))/(2*M_LN2));
	return res;
    }


    // sufficient_subdiv_level: NumType -> unsigned    
    // Parameter:
    //    epsilon   the tolerance, dafault value 0.001
    // Returns:  the number subdivision levels requiered to anssure that
    //           every segment of this curve will be approximated to within 
    //           given tolerance epsilon 
    unsigned sufficient_subdiv_level(NumType epsilon=EPSILON) const {
	if(S_curve::sufficient_subdiv_level(epsilon)==0) return 0;
	unsigned d = degree();
	assert(d>0);
	//L>=log4(d*(d-1)*M/8e)
	double arg =((d*(d-1))/_to_float((8*epsilon)))*max_second_diff_value();
	//log4(arg)=ln(arg)/ln(4)=ln(arg)/2*ln(2)=ln(arg)/2*M_LN2	
	unsigned res = (unsigned) ceil((log(arg))/(2*M_LN2));
	return res;
    }

    // wang_rendering: NumType -> Points
    // Parameters:
    //   epsilon   the tolerance
    //   points    vector of points wich approximates this curve within 
    //             a tolerance epsilon
    // Uses the Wang's Formulas
    // Ist rendering nicht gerade lin_approximation????
    void wang_rendering(Points& points, NumType epsilon=EPSILON) const {	
	unsigned levels = max_subdiv_level(epsilon);	
	//after level rounds subdivision there are 2^levels = 2<<levels-1 
	//points
	int number = 1<<levels;
	points.clear();
	points.resize(number+1);
	points[0] = contr_points[0];
	points[number] = contr_points[degree()];
	list<Bezier_curve_2> l;
	list<Bezier_curve_2>::iterator it;
	l.push_back(* this);
	int count, base, max_count;
	for(unsigned i=1; i<=levels; i++) {
	    max_count = 1<<(i-1);
	    it = l.begin(); 	    
	    count = 0;
	    base = 1<<(levels-i);
	    while(count<max_count/*it!=l.end()*/) {
		(*it++).subdivide(left_pts, right_pts);
		points[base*(1 + count*2)] = right_pts[0];
		Bezier_curve_2 right(right_pts);
		Bezier_curve_2 left(left_pts);
		l.push_back(left);				
		l.push_back(right);	
		l.pop_front();				
		count++;
	    }
	}	
    }
    
    // update: PointList -> SegmentList
    // Parameters:
    //    in     container with the control points of this Bezier curve
    //    out    container with the segments for drawing this Bezier curve
    // Effect: calls by every change of the input scene (the control points)
    //         and calculates the segments of the output scene (the curve)
    // Overwrites geowin_update<PointList, SegmentList>::update()
    void update (const PointList& in, SegmentList& out) {	
	out.clear();
	if(in.size() < 2) return;				
	set_control_points(in);	       			
	Points pts;
	rendering(pts);
	for(unsigned i = 1; i<pts.size(); i++) {
	    Segment s(pts[i-1], pts[i]);
	    out.push_back(s);
	}		
    }

    // draw: void -> void 
    // Effect: opens a GeoWin window and draws this Bezier curve
    // Overwrites Parametric_curve_2::draw()
    void draw() {
	geowin_init_default_type((PointList*)0, leda_string("ContrPointList"));
	geowin_init_default_type((SegmentList*)0, leda_string("SegmentList"));	
	GeoWin GW("Bezier Curves Demo");	
	GW.set_action( A_LEFT | A_DRAG | A_OBJECT, my_geo_object_dragging);
	
	//Stuetzpunkte in eine Liste kopieren
	PointList in_list;	
	in_list.resize(contr_points.size());
	copy(contr_points.begin(), contr_points.end(), in_list.begin());

	//the input scene is the scene with the control points
	geo_scene in_scene = GW.new_scene(in_list);
	GW.set_point_style(in_scene, leda_disc_point);
	GW.set_color(in_scene, leda_red);
	GW.set_visible(in_scene, true);
	GW.activate(in_scene);

	//result scene for the control polygon	
	geo_contr_polygon contr_polygon;
	geo_scene polygon_res_scene = GW.new_scene(contr_polygon, 
		          in_scene, leda_string("Control Polygon"));
	GW.set_color(polygon_res_scene, leda_red);
	GW.set_line_width(polygon_res_scene, 1);
	GW.set_visible(polygon_res_scene, true);	
	
	P_curve::draw(GW, in_scene, false);
	GW.edit(in_scene); 
    }    

    // bbox: void -> Bbox    
    // Returns: the bounding box of this Bezier curve wich is the bbox of the
    //          control points
    // Overwrites P_curve::bbox()
    Bbox bbox() const {	
	Lin_approximant tmp(contr_points.begin(), contr_points.end(), 
			    a, b);
	return tmp.bbox();
    };    

    // refine_approximantion: NumType -> void
    // Parameters:
    //    epsilon    for all t in [a,b] distatnce(appr(t), curve(t))<=epsilon,
    //               default value EPSILON
    // Effect: refines the piecewise linear approximant of this curve so that
    //         for all t in [a,b] distatnce(appr(t), curve(t))<=epsilon
    // Overwrites P_curve::refine_approximation()
    void refine_approximation(NumType epsilon=EPSILON, bool use_deriv=true) const {
	Points pts;
	if(use_deriv) {
	    unsigned r = sufficient_subdiv_level(epsilon);
	    rendering_n_times(pts, r);
	}
	else
	    rendering(pts, epsilon);
	ParamVector param;
	uniform_parametrisation(param, a, b, pts.size());
	lin_approx.set_points(pts.begin(), pts.end(), 
			      param.begin());
	eps = epsilon;
    }    


    // has_on: Point, NumType -> bool
    // Parameters: 
    //     pt       the point is tested if it is on the curve
    //     epsilon  the upper bound for the distance from pt to the 
    //              approximation of the curve
    // Returns: true, if pt lies on the curve (distance(pt, curve)<=epsilon)
    //          false otherwise
    // Tests if the point pt lies on the curve.
    // This function just solves the ambiguous problem with P_curve::has_on()
    // and S_curve::has_on() (Bezier inherits from both)
    bool has_on(const Point& pt, NumType epsilon=EPSILON) const {
	return P_curve::has_on(pt, epsilon);
    }

private:
    // the algorithm which calculates for a give parameter value t a point
    //on the Bezier curve
    A alg;

    // auxilary variables for the contruction
    Bezier_x bezier_x;
    Bezier_y bezier_y;   

    //auxiliary variables for some recurces
    mutable Points left_pts, right_pts;                   
    
};

template<class NumType> ostream & operator<< (ostream & os, Bezier_curve_2<NumType> & bz) {   
    Bezier_curve_2<NumType>::Points cntr_pt = bz.control_points();
    os<<"Stuetzpunkte der Bezier-Kurve: "<<endl;
    for(int i=0; i<=bz.degree(); i++) {
	os<<cntr_pt[i]<<" ";
    }
    os<<endl;
    return os;
}; 
#endif //CGAL_BEZIER_CURVE_2_H
