#ifndef CGAL_SUBDIVISION_CURVE_2_H
#define CGAL_SUBDIVISION_CURVE_2_H

#include "Types.h"
#include "Helpers.h"
#include "Curve_constants.h"
#include "Curve_2.h"


// Template class Subdivision_curve_2 implements a generalized Subdivision
// curve in 2D, defined with its control points.
// Template parameters:
//       NumType_      number type for coordinates
// Inherits from the class Curve_2.

template <class NumType_ > 
class Subdivision_curve_2 : public Curve_2<NumType_> {

public:  

    // Class geo_contr_polygon implements the drawing function of the control 
    // polygon of a subdivision curve
    // Inherits from geowin_update<PointList, SegmentList>
    class geo_contr_polygon : public geowin_update<PointList, SegmentList> {
	// update: PointList -> SegmentList
	// Parameters:
	//    in     container with the control points of the subdivision curve
	//    out    container with the segments of the control polygon of 
	//           this subdivision curve
	// Effect: calls by every change of the control points and links them
	//         to a control polygon
	// Overwrites geowin_update PointList, SegmentList>::update()
	void update (const PointList& in, SegmentList& out) {
	    link_points_to_segments(in, out);	   
	}
    };      

    // Subdivision_curve_2: void -> Subdivision_curve_2    
    // Effect: contructs a empty Subdivision_curve_2 object
    // contructor without parameter    
    Subdivision_curve_2(): contr_points() {}

    // Subdivision_curve_2: Points -> Subdivision_curve_2 
    // Parameters:
    //    contr_pts   the control points of the subdivision curve   
    // Effect: contructs a Subdivision_curve_2 object whit control 
    // points contr_pts
    Subdivision_curve_2(Points contr_pts): 
	contr_points(contr_pts) {}

    // Subdivision_curve_2: Subdivision_curve_2 -> Subdivision_curve_2 
    // Parameters:
    //    curve   Subdivision curve
    // Effect: contructs a Subdivision_curve_2 object which is a 
    //         copy of curve   
    Subdivision_curve_2(const Subdivision_curve_2 & curve) : 
    	contr_points(curve.control_points()) {}

    // Destructor
    virtual ~Subdivision_curve_2() {
    }

    // degree: void -> unsigned    
    // Pre:    the curve has at least one control point
    // Returns the degree of this Subdivision curve
    virtual unsigned degree() const {
	assert(!contr_points.empty());       
	return (contr_points.size()-1); 	
    };

    // initial_epsilon: void -> NumType
    // Returns: 
    //   a initial value for epsilon wich is the largest distance 
    //   from a internal control point to the segment joining the first and 
    //   the last control points of this subdivision curve
    NumType initial_epsilon() const {
	if(!contr_points.empty()) {
	    unsigned n = contr_points.size()-1;	    
	    Segment s(contr_points[0], contr_points[n]);
	    NumType max_dist = 0;
	    NumType dist;	
	    for(unsigned i=1; i<n; i++) {
		dist = squared_distance(contr_points[i], s);		
		if(dist > max_dist) max_dist = dist;
	    }    	    
	    return _sqrt(max_dist);
	}
	else return 1000.0;	
    }

    // control_points: void -> Points
    // Returns: the control points of this subdivision curve
    Points control_points() const {
       return contr_points;
    };

    // set_control_points: PointList -> void
    // Parameters:
    //    new_contr_pts   the new control points of this Subdivision curve
    // Effect:  sets the control points of this subdivision curve to 
    //          new_contr_pts
    void set_control_points(const PointList& new_contr_pts) {
	contr_points.resize(new_contr_pts.size());
	copy(new_contr_pts.begin(),new_contr_pts.end(), contr_points.begin());	
    };

    
    // discrete_differences: unsigned -> Points
    // Papameters:    
    //    r          compute the r-th control point differences of 
    //               this Subdivision curve, default value r=1
    //    diff       contains after the execution the r-th control point 
    //               diferences    
    // Pre:     0< r <=degree()    
    virtual bool discrete_differences(Points & diff, unsigned r=1) const {
        unsigned n = contr_points.size()-1;
        assert(r > 0);
	assert(n>=r); 
	unsigned d = degree();      
	assert(d>=r);        	
	diff = control_points();
	for(unsigned j = (n-1); j>=(n-r); j--) {
	    for(unsigned k = 0; k<=j; k++) { 
		diff[k] = CGAL::ORIGIN + (diff[k+1]-diff[k]);
	    }
	    //dieser break is noetig weil unsigned j=0-1=49...90 >0
	    //und die schleife geht ewig.
	    if(j==0) break; 
	}	
	diff.resize(n-r+1);	
    }
   
    // max_second_diff_value: void -> NumType    
    // Returns:  Max|P_k+2 -2*P_k+1 + P_k| - the value of the maximal 
    //           second control point difference
    double max_second_diff_value() const {
	if(degree()<=1) return 0;	
	Points pts;
	discrete_differences(pts, 2);
	//max_2= (Max|P_k+2 - 2*P_k+1 + P_k|)^2
	NumType max_2 = (pts[0] - CGAL::ORIGIN)*(pts[0] - CGAL::ORIGIN);
	NumType tmp_2;
	for(unsigned i=1; i<pts.size(); i++) {
	    tmp_2 = (pts[i] - CGAL::ORIGIN)*(pts[i] - CGAL::ORIGIN);	    
	    if(max_2<tmp_2) max_2=tmp_2;
	}	
	return _sqrt(max_2);
    }

    // sufficient_subdiv_level: NumType -> unsigned    
    // Parameter:
    //    epsilon   the tolerance, dafault value 0.001
    // Returns:  the number subdivision levels requiered to anssure that
    //           every segment of this curve will be approximated to within 
    //           given tolerance epsilon 
    // Must be overwriten of the derived classes
    // Po vsiaka veroiatnost triabva da stane pure virtual toia metod
    virtual unsigned sufficient_subdiv_level(NumType epsilon=EPSILON) const {
	//there is no need of subdivision for 0 or 1 control points
	if(contr_points.empty() || contr_points.size()<=1) return 0;	
	
	return 3;
    }
       

    // draw: void -> void 
    // Effect: opens a GeoWin window and draws this curve
    // Overwrites Curve_2::draw()
    virtual void draw() {}

    // has_on: Point, NumType -> bool
    // Parameters: 
    //     pt       the point is tested if it is on the curve
    //     epsilon  the upper bound for the distance from pt to the 
    //              approximation of the curve
    // Returns: true, if pt lies on the curve (distance(pt, curve)<=epsilon)
    //          false otherwise
    // Tests if the point pt lies on the curve.
    bool has_on(const Point& pt, NumType epsilon=EPSILON) const {
	return true;
    }

protected:
    //the control points of this subdivision curve
    Points contr_points;
    
};
#endif //CGAL_SUBDIVISION_CURVE_2_H

